"""
Unit tests for the otter.Otter class methods
"""

import os
from otter import Otter, Transient
from otter.exceptions import FailedQueryError
from astropy.coordinates import SkyCoord
from astropy.table import Table
import numpy as np
import pandas as pd
import pytest


def test_otter_constructor():
    """
    Just make sure everything constructs correctly
    """

    db = Otter()
    assert isinstance(db, Otter)

    # and using a custom local directory
    otter_test_dir = os.path.join(os.getcwd(), "otter-test")
    db = Otter(
        datadir=otter_test_dir,
        gen_summary=True,
    )

    assert os.path.exists(otter_test_dir)  # make sure it was created
    assert isinstance(db, Otter)


def test_get_meta():
    """
    Tests the Otter.get_meta method and make sure it returns as expected
    """

    db = Otter()

    # first make sure everything is just copied over correctly
    allmeta = db.get_meta()
    req_keys = ["name", "coordinate"]

    assert all(k in d for d in allmeta for k in req_keys)

    # now we can try real queries
    true_keys = ["name", "coordinate", "date_reference", "distance", "classification"]
    metahyz = db.get_meta(names="2018hyz")[0]
    assert isinstance(metahyz, Transient)
    assert all(k in metahyz for k in true_keys)
    assert metahyz["name/default_name"] == "2018hyz"
    assert metahyz["date_reference"][0]["value"] == "2018-10-14"
    assert metahyz["date_reference"][0]["date_format"] == "iso"
    assert metahyz["classification"]["value"][0]["object_class"] == "TDE"


def test_cone_search():
    """
    Tests the Otter.cone_search method
    """

    db = Otter()

    # just search around '2018hyz' coordinates to make sure it picks it up
    coord = SkyCoord(151.711964138, 1.69279894089, unit="deg")
    res = db.cone_search(coord)[0]
    assert res["name/default_name"] == "2018hyz"


def test_get_phot():
    """
    Tests the Otter.get_phot method

    We know from the transients.clean_photometry tests that the conversions
    work as expected. So, this will just test that everything comes out as expected.
    """

    db = Otter()

    true_keys = [
        "name",
        "converted_flux",
        "converted_flux_err",
        "converted_date",
        "converted_wave",
        "converted_freq",
        "converted_flux_unit",
        "converted_date_unit",
        "converted_wave_unit",
        "converted_freq_unit",
        "obs_type",
        "upperlimit",
        "reference",
    ]

    names = ["2018hyz", "2018zr", "ASASSN-14li", "2016fnl", "J123715"]

    # first with returning an astropy table (the default)
    allphot = db.get_phot(names=names)
    assert isinstance(allphot, Table)
    assert all(k in allphot.keys() for k in true_keys)
    assert len(np.unique(allphot["converted_flux_unit"])) == 1
    assert allphot["converted_flux_unit"][0] == "mag(AB)"

    # then with returning a pandas DataFrame
    allphot = db.get_phot(names=names, return_type="pandas")
    assert isinstance(allphot, pd.DataFrame)
    assert all(k in allphot for k in true_keys)
    assert len(np.unique(allphot.converted_flux_unit)) == 1
    assert allphot.converted_flux_unit.iloc[0] == "mag(AB)"

    # then make sure it throws the FailedQueryError
    with pytest.raises(FailedQueryError):
        db.get_phot(names="foo")

    # some other random tests
    with pytest.raises(OSError):
        db.get_phot(names="ASASSN-14li", return_type="foobar")


def test_query():
    """
    Tests the Otter.query method that basically all of this is based on

    A lot of these have been tested in other unit tests in thie file
    but lets make sure it's complete
    """

    db = Otter()

    # test min and max z queries
    zgtr1 = db.query(minz=1)
    assert len(zgtr1) >= 2
    true_result = ["Swift J2058.4+0516", "2022cmc", "CXOU J0332"]
    assert all(t["name/default_name"] in true_result for t in zgtr1)

    zless001 = db.query(maxz=0.001)
    result = ["NGC 247", "IGR J17361-4441"]
    assert all(t["name/default_name"] in result for t in zless001)

    # test refs
    # res = db.query(refs="2020MNRAS.tmp.2047S")[0]
    # assert res["name/default_name"] == "2018hyz"

    # test hasphot and hasspec
    assert len(db.query(hasspec=True)) == 0
    assert "ASASSN-20il" not in {t["name/default_name"] for t in db.query(hasphot=True)}

    # test has_*_phot
    assert len(db.query(has_radio_phot=True)) >= 92
    assert len(db.query(has_xray_phot=True)) >= 35
    assert len(db.query(has_uvoir_phot=True)) >= 120
    assert len(db.query(has_radio_phot=True, has_xray_phot=True)) < len(
        db.query(has_radio_phot=True)
    )

    # test classification related queries
    assert len(db.query(spec_classed=True)) > 140
    assert len(db.query(unambiguous=True)) > 190
    assert len(db.query(classification="SLSN")) >= 1

    # check that querying based on references works
    fake_test_bibcodes = ["'tasdfasdf...lkjsfd'", "'asfd...kjasdf...lkjs'"]
    assert len(db.query(refs=fake_test_bibcodes)) == 0
    assert len(db.query(refs=fake_test_bibcodes[0])) == 0

    # try querying with the fake "private" data
    db2 = Otter(
        datadir=os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "private_otter_data"
        ),
        gen_summary=True,
    )
    res = db2.query(names="2018hyz", query_private=True)
    assert res[0].default_name == "2018hyz"

    # check that some errors are thrown when appropriate
    with pytest.raises(Exception):
        db.query(names=10)
        db.query(refs=10)

    # try just retrieving all of the private data
    all_res = db2._query_datadir()
    assert len(all_res) == 1

    # check some other basic queries agains the private data
    res = db2.query(names=["2018hyz", "2022cmc"])
    assert len(res) == 2

    with pytest.raises(AttributeError):
        res = db2.query(names="2018hyz", hasspec=True, query_private=True)

    res = db2.query(names="2018hyz", hasphot=True, query_private=True)
    assert len(res) == 1

    res = db2.query(names="2018hyz", maxz=1, query_private=True)
    assert len(res) == 1

    res = db2.query(minz=1, query_private=True)
    assert len(res) >= 3
    assert "2018hyz" not in [t.default_name for t in res]

    res = db2._query_datadir(refs="2018TNSCR1764....1A")
    assert len(res) == 1

    res = db2._query_datadir(refs=["2018TNSCR1764....1A", "2018TNSTR1708....1B"])
    assert len(res) == 1

    # THESE ALL MUST HAPPEN AT THE END OF THE QUERYING
    # they are testing what happens if we mess with the summary table in the
    # private data directory
    summary_table_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "private_otter_data", "summary.csv"
    )
    summary_table = pd.read_csv(summary_table_path)
    summary_table.head(0).to_csv(summary_table_path)
    res = db2._query_datadir(names="2018hyz")
    assert isinstance(res, list)
    assert len(res) == 0

    # now delete the summary file because it should auto regenerate
    os.remove(summary_table_path)
    res = db2._query_datadir(names="2018hyz")
    assert len(res) == 1


def test_from_csvs():
    """
    This tests the "from_csvs" method which allows for interaction with locally
    stored datasets
    """

    metapath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "meta-test.csv")
    photpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "phot-test.csv")
    outpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "from-csvs-out")

    db = Otter.from_csvs(metafile=metapath, photfile=photpath, local_outpath=outpath)

    assert isinstance(db, Otter)

    metapath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "gaia16aax-meta.csv"
    )
    photpath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "gaia16aax-phot.csv"
    )
    outpath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "from-csvs-out-gaia16aax"
    )

    db = Otter.from_csvs(metafile=metapath, photfile=photpath, local_outpath=outpath)

    assert isinstance(db, Otter)



"""
Some testing for the mini plotting module that's part of this api

Author: Noah Franz
"""

import os
import matplotlib.pyplot as plt
import plotly
from otter import Otter
from otter.plotter import plotter
import pytest

OTTER_URL = os.environ.get("OTTER_TEST_URL")
OTTER_TEST_PASSWORD = os.environ.get("OTTER_TEST_PASSWORD")


def _get_test_data():
    db = Otter(url=OTTER_URL, password=OTTER_TEST_PASSWORD)
    t = db.query(names="ASASSN-14li")
    return t[0]


def test_otter_plotter_class():
    # try initializing the class a bunch of ways
    op = plotter.OtterPlotter(backend="matplotlib.pyplot")
    assert op.backend == "matplotlib.pyplot"

    op = plotter.OtterPlotter(backend="plotly.graph_objects")
    assert op.backend == "plotly.graph_objects"

    op = plotter.OtterPlotter(backend="plotly")
    assert op.backend == "plotly.graph_objects"

    op = plotter.OtterPlotter(backend="matplotlib")
    assert op.backend == "matplotlib.pyplot"

    with pytest.raises(ValueError):
        plotter.OtterPlotter(backend="test")

    # then test the two backends with some basics
    op = plotter.OtterPlotter(backend="matplotlib")
    ax = op.plot([1], [1])
    assert isinstance(ax, plt.Axes)

    op = plotter.OtterPlotter(backend="plotly")
    fig = op.plot([1], [1])
    assert isinstance(fig, plotly.graph_objs._figure.Figure)


def test_plot_light_curve():
    t = _get_test_data()
    phot = t.clean_photometry(obs_type="uvoir", flux_unit="mag(AB)")

    # first try with matplotlib
    ax = plotter.plot_light_curve(
        phot.converted_date,
        phot.converted_flux,
        date_err=None,
        flux_err=phot.converted_flux_err,
        backend="matplotlib",
    )
    assert isinstance(ax, plt.Axes)

    # then try with plotly
    fig = plotter.plot_light_curve(
        phot.converted_date,
        phot.converted_flux,
        date_err=None,
        flux_err=phot.converted_flux_err,
        backend="plotly",
    )
    assert isinstance(fig, plotly.graph_objs._figure.Figure)


def test_plot_sed():
    t = _get_test_data()
    phot = t.clean_photometry(obs_type="uvoir", flux_unit="mag(AB)")

    ax = plotter.plot_sed(
        phot.converted_wave,
        phot.converted_flux,
        flux_err=phot.converted_flux_err,
        backend="matplotlib",
        ylabel="Flux [AB]",
        xlabel="Wavelength [nm]",
    )
    assert isinstance(ax, plt.Axes)
    assert ax.get_ylabel() == "Flux [AB]"
    assert ax.get_xlabel() == "Wavelength [nm]"

    fig = plotter.plot_sed(
        phot.converted_freq,
        phot.converted_flux,
        flux_err=phot.converted_flux_err,
        backend="plotly",
        ylabel="Flux [AB]",
        xlabel="Frequency [GHz]",
    )
    assert isinstance(fig, plotly.graph_objs._figure.Figure)
    assert fig.layout.xaxis.title.text == "Frequency [GHz]"
    assert fig.layout.yaxis.title.text == "Flux [AB]"


def test_quick_view():
    t = _get_test_data()

    fig = plotter.quick_view(t, ptype="both", sed_dim="wave", obs_type="uvoir")

    assert isinstance(fig, plt.Figure)
    assert len(fig.get_axes()) == 2
    lc_ax, sed_ax = fig.get_axes()
    assert lc_ax.get_xlabel() == "Date [MJD]"
    assert lc_ax.get_ylabel() == "Flux [mag(AB)]"
    assert sed_ax.get_xlabel() == "Wavelength [nm]"
    assert sed_ax.get_ylabel() == "Flux [mag(AB)]"

    fig = plotter.quick_view(t, ptype="lc", obs_type="radio", flux_unit="mJy")
    assert isinstance(fig, plt.Figure)
    assert len(fig.get_axes()) == 1
    ax = fig.get_axes()[0]
    assert ax.get_xlabel() == "Date [MJD]"
    assert ax.get_ylabel() == "Flux [mJy]"

    fig = plotter.quick_view(t, ptype="sed", obs_type="uvoir")
    assert isinstance(fig, plt.Figure)
    assert len(fig.get_axes()) == 1
    ax = fig.get_axes()[0]
    assert ax.get_xlabel() == "Frequency [GHz]"
    assert ax.get_ylabel() == "Flux [mag(AB)]"

    with pytest.raises(ValueError):
        plotter.quick_view(t, plotting_kwargs=dict(backend="plotly"))


def test_query_quick_view():
    db = Otter(url=OTTER_URL, password=OTTER_TEST_PASSWORD)

    fig = plotter.query_quick_view(
        db,
        ptype="both",
        sed_dim="wave",
        phot_cleaning_kwargs=dict(obs_type="uvoir"),
        names="ASASSN-14li",
    )

    assert all(isinstance(f, plt.Figure) for f in fig)
    assert len(fig[0].get_axes()) == 2
    lc_ax, sed_ax = fig[0].get_axes()
    assert lc_ax.get_xlabel() == "Date [MJD]"
    assert lc_ax.get_ylabel() == "Flux [mag(AB)]"
    assert sed_ax.get_xlabel() == "Wavelength [nm]"
    assert sed_ax.get_ylabel() == "Flux [mag(AB)]"

    fig = plotter.query_quick_view(
        db,
        ptype="lc",
        phot_cleaning_kwargs=dict(obs_type="radio", flux_unit="mJy"),
        names="Sw J1644+57",
    )
    assert isinstance(fig[0], plt.Figure)
    assert len(fig[0].get_axes()) == 1
    ax = fig[0].get_axes()[0]
    assert ax.get_xlabel() == "Date [MJD]"
    assert ax.get_ylabel() == "Flux [mJy]"

    fig = plotter.query_quick_view(
        db,
        ptype="sed",
        phot_cleaning_kwargs=dict(obs_type="uvoir"),
        names="ASASSN-14li",
    )
    assert isinstance(fig[0], plt.Figure)
    assert len(fig[0].get_axes()) == 1
    ax = fig[0].get_axes()[0]
    assert ax.get_xlabel() == "Frequency [GHz]"
    assert ax.get_ylabel() == "Flux [mag(AB)]"

    with pytest.raises(RuntimeError):
        fig = plotter.query_quick_view(db, minz=1, maxz=2, result_length_tol=1)



"""
Test the utility functions used throughout the OTTER API
"""

import json
from copy import deepcopy
from otter import util
from astropy import units as u
import pytest


def test_filter_to_obstype():
    """
    Test the converion from a filtername to a obstype variable. This is used a lot
    during the data cleaning process
    """

    assert util.filter_to_obstype("FUV") == "uvoir"
    assert util.filter_to_obstype("r") == "uvoir"

    with pytest.raises(Exception):
        util.filter_to_obstype("some fake filter name")


def test_freq_to_obstype():
    """
    Test the frequency conversion to an obstype
    """
    assert util.freq_to_obstype(1 * u.GHz) == "radio"


def test_wave_to_obstype():
    assert util.wave_to_obstype(100 * u.nm) == "uvoir"
    assert util.wave_to_obstype(1 * u.m) == "radio"
    assert util.wave_to_obstype(1e-3 * u.nm) == "xray"


def test_clean_schema():
    """
    Also used a lot during the data cleaning process. This function tests the
    clean_schema method.
    """

    subschema = deepcopy(util.distance_schema)

    subschema["value"] = "foo"
    subschema["reference"] = "bar"

    cleaned_schema = util.clean_schema(subschema)
    assert "unit" not in cleaned_schema
    assert "value" in cleaned_schema
    assert "reference" in cleaned_schema
    assert "error" not in cleaned_schema
    assert cleaned_schema["value"] == "foo"
    assert cleaned_schema["reference"] == "bar"


def test_bibcode_to_hrn():
    """
    Test the bibcode to hrn that only relies on a local file
    """
    bibcode_to_hrn = {"foo": "bar", "xyz": "wqr"}

    with open("reference_map_local.json", "w") as f:
        json.dump(bibcode_to_hrn, f, indent=4)

    # now we can actually test the local version
    bibcodes, hrns = util.bibcode_to_hrn("foo")
    assert bibcodes == ["foo"]
    assert hrns == ["bar"]

    bibcodes, hrns = util.bibcode_to_hrn(["foo", "xyz"])
    assert bibcodes == ["foo", "xyz"]
    assert hrns == ["bar", "wqr"]


def test_freq_to_band():
    """
    Test the conversion from a frequency quantity to a single band
    """

    res = util.freq_to_band(1.4 * u.GHz)
    assert res == "L"

    res = util.freq_to_band(10 * u.GHz)
    assert res == "X"

    with pytest.raises(ValueError):
        res = util.freq_to_band(1 * u.Hz)


def test_freqlist_to_band():
    res = util.freqlist_to_band([1, 10], ["GHz", "GHz"])
    assert res[0] == "UHF" and res[1] == "X"



"""
Test the schema definitions and validators
"""

from otter.schema import (
    CoordinateSchema,
    DistanceSchema,
    PhotometrySchema,
    HostSchema,
    OtterSchema,
)
from pydantic import ValidationError
import pytest


def test_coordinate_validator():
    # first, some that should succeed
    case1 = dict(reference="foo", ra=1, dec=1, ra_units="deg", dec_units="deg")
    cs1 = CoordinateSchema(**case1)
    assert cs1.ra == 1 and cs1.dec == 1

    case2 = dict(reference="foo", l=1, b=1, l_units="deg", b_units="deg")
    cs1 = CoordinateSchema(**case2)
    assert cs1.l == 1 and cs1.b == 1

    case3 = dict(reference="foo", lon=1, lat=1, lon_units="deg", lat_units="deg")
    cs1 = CoordinateSchema(**case3)
    assert cs1.lon == 1 and cs1.lat == 1

    # now some cases that won't work
    del case1["dec_units"]
    with pytest.raises(ValueError):
        CoordinateSchema(**case1)

    del case1["ra_units"]
    with pytest.raises(ValueError):
        CoordinateSchema(**case1)

    del case2["b_units"]
    with pytest.raises(ValueError):
        CoordinateSchema(**case2)

    del case2["l_units"]
    with pytest.raises(ValueError):
        CoordinateSchema(**case2)

    del case3["lat_units"]
    with pytest.raises(ValueError):
        CoordinateSchema(**case3)

    del case3["lon_units"]
    with pytest.raises(ValueError):
        CoordinateSchema(**case3)

    case4 = {"reference": "foo"}
    with pytest.raises(ValueError):
        CoordinateSchema(**case4)


def test_distance_validator():
    case1 = dict(value=1, reference="foo", distance_type="redshift")

    case2 = dict(value=1, reference="foo", distance_type="luminosity", unit="cm")

    cs1 = DistanceSchema(**case1)
    assert cs1.value == 1
    assert cs1.reference == "foo"
    assert cs1.unit is None

    cs2 = DistanceSchema(**case2)
    assert cs2.value == 1
    assert cs2.reference == "foo"
    assert cs2.distance_type == "luminosity"
    assert cs2.unit == "cm"

    del case2["unit"]
    with pytest.raises(ValueError):
        DistanceSchema(**case2)


def test_photometry_validators():
    case1 = dict(
        reference="foo",
        raw=[1, 2],
        raw_err=[3, 4],
        raw_units="mJy",
        filter_key="X",
        obs_type="radio",
        date=[1, 2],
        date_format="mjd",
    )

    PhotometrySchema(**case1)

    case2 = dict(
        reference="foo",
        raw=1,
        raw_err=3,
        raw_units="mJy",
        filter_key="X",
        obs_type="radio",
        date=1,
        date_format="mjd",
    )

    with pytest.raises(ValueError):
        PhotometrySchema(**case2)


def test_host_validator():
    case1 = dict(reference="foo", ra=1, dec=1, ra_units="deg")

    with pytest.raises(ValueError):
        HostSchema(**case1)

    del case1["ra_units"]
    with pytest.raises(ValueError):
        HostSchema(**case1)

    case2 = dict(reference="foo", host_ra_units="deg", host_dec_units="deg")

    with pytest.raises(ValueError):
        HostSchema(**case2)

    case3 = dict(
        reference="foo",
        host_ra=1,
        host_ra_units="deg",
        host_dec_units="deg",
        host_name="foo",
    )

    with pytest.raises(ValueError):
        HostSchema(**case3)

    case4 = dict(
        reference="foo",
        host_ra=1,
        host_dec=1,
        host_ra_units="deg",
        host_dec_units="deg",
        host_name="foo",
    )

    case5 = dict(reference="foo", host_name="foo")

    case6 = dict(
        host_ra=1,
        host_dec=1,
        host_ra_units="deg",
        host_dec_units="deg",
    )

    HostSchema(**case4)
    HostSchema(**case5)

    with pytest.raises(ValidationError):
        HostSchema(**case6)


def test_otter_validator():
    case1 = dict(
        name=dict(default_name="foo", alias=[dict(value="foo", reference="foo")]),
        coordinate=[
            dict(ra=1, dec=1, ra_units="deg", dec_units="deg", reference="foo")
        ],
        photometry=[
            dict(
                reference="foo",
                raw=[1, 2],
                raw_err=[3, 4],
                raw_units="mJy",
                filter_key="X",
                obs_type="radio",
                date=[1, 2],
                date_format="mjd",
            )
        ],
        reference_alias=[{"name": "foo", "human_readable_name": "foo"}],
    )

    with pytest.raises(ValueError):
        OtterSchema(**case1)

    case1["filter_alias"] = [
        dict(filter_key="X", filter_name="X", wave_eff=1, wave_units="nm")
    ]

    OtterSchema(**case1)
